@@ -1,5 +1,5 @@
 # This file is part of pyacoustid.
-# Copyright 2012, Adrian Sampson.
+# Copyright 2014, Adrian Sampson.
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -8,15 +8,16 @@
 # distribute, sublicense, and/or sell copies of the Software, and to
 # permit persons to whom the Software is furnished to do so, subject to
 # the following conditions:
-# 
+#
 # The above copyright notice and this permission notice shall be
 # included in all copies or substantial portions of the Software.
 
+from __future__ import division
+from __future__ import absolute_import
+
 import os
 import json
-import urllib
-import urllib2
-import httplib
+import requests
 import contextlib
 import errno
 try:
@@ -33,7 +34,8 @@ import subprocess
 import threading
 import time
 import gzip
-from StringIO import StringIO
+from io import BytesIO
+
 
 API_BASE_URL = 'http://api.acoustid.org/v2/'
 DEFAULT_META = 'recordings'
@@ -42,20 +44,27 @@ MAX_AUDIO_LENGTH = 120 # Seconds.
 FPCALC_COMMAND = 'fpcalc'
 FPCALC_ENVVAR = 'FPCALC'
 
+
+# Exceptions.
+
 class AcoustidError(Exception):
     """Base for exceptions in this module."""
 
+
 class FingerprintGenerationError(AcoustidError):
     """The audio could not be fingerprinted."""
 
+
 class NoBackendError(FingerprintGenerationError):
     """The audio could not be fingerprinted because neither the
     Chromaprint library nor the fpcalc command-line tool is installed.
     """
 
+
 class FingerprintSubmissionError(AcoustidError):
     """Missing required data for a fingerprint submission."""
 
+
 class WebServiceError(AcoustidError):
     """The Web service request failed. The field ``message`` contains a
     description of the error. If this is an error that was specifically
@@ -83,6 +92,51 @@ class WebServiceError(AcoustidError):
         super(WebServiceError, self).__init__(message)
         self.message = message
 
+
+# Endpoint configuration.
+
+def set_base_url(url):
+    """Set the URL of the API server to query."""
+    if not url.endswith('/'):
+        url += '/'
+    global API_BASE_URL
+    API_BASE_URL = url
+
+
+def _get_lookup_url():
+    """Get the URL of the lookup API endpoint."""
+    return API_BASE_URL + 'lookup'
+
+
+def _get_submit_url():
+    """Get the URL of the submission API endpoint."""
+    return API_BASE_URL + 'submit'
+
+
+# Compressed HTTP request bodies.
+
+def _compress(data):
+    """Compress a bytestring to a gzip archive."""
+    sio = BytesIO()
+    with contextlib.closing(gzip.GzipFile(fileobj=sio, mode='wb')) as f:
+        f.write(data)
+    return sio.getvalue()
+
+
+class CompressedHTTPAdapter(requests.adapters.HTTPAdapter):
+    """An `HTTPAdapter` that compresses request bodies with gzip. The
+    Content-Encoding header is set accordingly.
+    """
+    def add_headers(self, request, **kwargs):
+        body = request.body
+        if not isinstance(body, bytes):
+            body = body.encode('utf8')
+        request.prepare_body(_compress(body), None)
+        request.headers['Content-Encoding'] = 'gzip'
+
+
+# Utilities.
+
 class _rate_limit(object):
     """A decorator that limits the rate at which the function may be
     called.  The rate is controlled by the REQUEST_INTERVAL module-level
@@ -107,80 +161,33 @@ class _rate_limit(object):
             # Call the original function.
             return self.fun(*args, **kwargs)
 
-def _compress(data):
-    """Compress a string to a gzip archive."""
-    sio = StringIO()
-    with contextlib.closing(gzip.GzipFile(fileobj=sio, mode='wb')) as f:
-        f.write(data)
-    return sio.getvalue()
-
-def _decompress(data):
-    """Decompress a gzip archive contained in a string."""
-    sio = StringIO(data)
-    with contextlib.closing(gzip.GzipFile(fileobj=sio)) as f:
-        return f.read()
-
-def set_base_url(url):
-    """Set the URL of the API server to query."""
-    if not url.endswith('/'):
-        url += '/'
-    global API_BASE_URL
-    API_BASE_URL = url
-
-def _get_lookup_url():
-    """Get the URL of the lookup API endpoint."""
-    return API_BASE_URL + 'lookup'
-
-def _get_submit_url():
-    """Get the URL of the submission API endpoint."""
-    return API_BASE_URL + 'submit'
 
 @_rate_limit
-def _send_request(req):
-    """Given a urllib2 Request object, make the request and return a
-    tuple containing the response data and headers.
-    """
-    try:
-        with contextlib.closing(urllib2.urlopen(req)) as f:
-            return f.read(), f.info()
-    except urllib2.HTTPError as exc:
-        raise WebServiceError('HTTP status %i' % exc.code, exc.read())
-    except httplib.BadStatusLine:
-        raise WebServiceError('bad HTTP status line')
-    except IOError:
-        raise WebServiceError('connection failed')
-
 def _api_request(url, params):
     """Makes a POST request for the URL with the given form parameters,
     which are encoded as compressed form data, and returns a parsed JSON
     response. May raise a WebServiceError if the request fails.
     """
-    # Encode any Unicode values in parameters. (urllib.urlencode in
-    # Python 2.x operates on bytestrings, so a Unicode error is raised
-    # if non-ASCII characters are passed in a Unicode string.)
-    byte_params = {}
-    for key, value in params.iteritems():
-        if isinstance(key, unicode):
-            key = key.encode('utf8')
-        if isinstance(value, unicode):
-            value = value.encode('utf8')
-        byte_params[key] = value
-
-    body = _compress(urllib.urlencode(byte_params))
-    req = urllib2.Request(url, body, {
-        'Content-Encoding': 'gzip',
+    headers = {
         'Accept-Encoding': 'gzip',
-    })
+        "Content-Type": "application/x-www-form-urlencoded"
+    }
 
-    data, headers = _send_request(req)
-    if headers.get('Content-Encoding') == 'gzip':
-        data = _decompress(data)
+    session = requests.Session()
+    session.mount('http://', CompressedHTTPAdapter())
+    try:
+        response = session.post(url, data=params, headers=headers)
+    except requests.exceptions.RequestException as exc:
+        raise WebServiceError("HTTP request failed: {0}".format(exc))
 
     try:
-        return json.loads(data)
+        return response.json()
     except ValueError:
         raise WebServiceError('response is not valid JSON')
 
+
+# Main API.
+
 def fingerprint(samplerate, channels, pcmiter, maxlength=MAX_AUDIO_LENGTH):
     """Fingerprint audio data given its sample rate and number of
     channels.  pcmiter should be an iterable containing blocks of PCM
@@ -205,6 +212,7 @@ def fingerprint(samplerate, channels, pcmiter, maxlength=MAX_AUDIO_LENGTH):
     except chromaprint.FingerprintError:
         raise FingerprintGenerationError("fingerprint calculation failed")
 
+
 def lookup(apikey, fingerprint, duration, meta=DEFAULT_META):
     """Look up a fingerprint with the Acoustid Web service. Returns the
     Python object reflecting the response JSON data.
@@ -218,6 +226,7 @@ def lookup(apikey, fingerprint, duration, meta=DEFAULT_META):
     }
     return _api_request(_get_lookup_url(), params)
 
+
 def parse_lookup_result(data):
     """Given a parsed JSON response, generate tuples containing the match
     score, the MusicBrainz recording ID, the title of the recording, and
@@ -246,6 +255,7 @@ def parse_lookup_result(data):
 
             yield score, recording['id'], recording.get('title'), artist_name
 
+
 def _fingerprint_file_audioread(path, maxlength):
     """Fingerprint a file by using audioread and chromaprint."""
     try:
@@ -256,6 +266,7 @@ def _fingerprint_file_audioread(path, maxlength):
         raise FingerprintGenerationError("audio could not be decoded")
     return duration, fp
 
+
 def _fingerprint_file_fpcalc(path, maxlength):
     """Fingerprint a file by calling the fpcalc application."""
     fpcalc = os.environ.get(FPCALC_ENVVAR, FPCALC_COMMAND)
@@ -284,21 +295,22 @@ def _fingerprint_file_fpcalc(path, maxlength):
     duration = fp = None
     for line in output.splitlines():
         try:
-            parts = line.split('=', 1)
+            parts = line.split(b'=', 1)
         except ValueError:
             raise FingerprintGenerationError("malformed fpcalc output")
-        if parts[0] == 'DURATION':
+        if parts[0] == b'DURATION':
             try:
                 duration = int(parts[1])
             except ValueError:
                 raise FingerprintGenerationError("fpcalc duration not numeric")
-        elif parts[0] == 'FINGERPRINT':
+        elif parts[0] == b'FINGERPRINT':
             fp = parts[1]
 
     if duration is None or fp is None:
         raise FingerprintGenerationError("missing fpcalc output")
     return duration, fp
 
+
 def fingerprint_file(path, maxlength=MAX_AUDIO_LENGTH):
     """Fingerprint a file either using the Chromaprint dynamic library
     or the fpcalc command-line tool, whichever is available. Returns the
@@ -310,6 +322,7 @@ def fingerprint_file(path, maxlength=MAX_AUDIO_LENGTH):
     else:
         return _fingerprint_file_fpcalc(path, maxlength)
 
+
 def match(apikey, path, meta=DEFAULT_META, parse=True):
     """Look up the metadata for an audio file. If ``parse`` is true,
     then ``parse_lookup_result`` is used to return an iterator over
@@ -323,6 +336,7 @@ def match(apikey, path, meta=DEFAULT_META, parse=True):
     else:
         return response
 
+
 def submit(apikey, userkey, data):
     """Submit a fingerprint to the acoustid server. The ``apikey`` and
     ``userkey`` parameters are API keys for the application and the
